@@ -17,10 +17,12 @@
 
 import abc
 import collections
+from concurrent import futures
 import enum
 import logging
 import threading
 
+from oslo_utils import encodeutils
 from oslo_utils import excutils
 from oslo_utils import netutils
 from oslo_utils import timeutils
@@ -28,6 +30,9 @@ import six
 from stevedore import driver
 
 import tooz
+from tooz import _retry
+from tooz import partitioner
+from tooz import utils
 
 LOG = logging.getLogger(__name__)
 
@@ -54,7 +59,7 @@ class Characteristics(enum.Enum):
     """The driver has the following property:
 
     * Its operations are not based on the timeout of other clients, but on some
-    other more robust mechanisms.
+      other more robust mechanisms.
     """
 
     LINEARIZABLE = 'LINEARIZABLE'
@@ -126,6 +131,11 @@ class MemberJoinedGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: +member %s>" % (self.__class__.__name__,
+                                               self.group_id,
+                                               self.member_id)
+
 
 class MemberLeftGroup(Event):
     """A member left a group event."""
@@ -134,6 +144,11 @@ class MemberLeftGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: -member %s>" % (self.__class__.__name__,
+                                               self.group_id,
+                                               self.member_id)
+
 
 class LeaderElected(Event):
     """A leader as been elected."""
@@ -150,8 +165,6 @@ class Heart(object):
                  event_cls=threading.Event):
         self._thread_cls = thread_cls
         self._dead = event_cls()
-        self._finished = event_cls()
-        self._finished.set()
         self._runner = None
         self._driver = driver
         self._beats = 0
@@ -164,38 +177,36 @@ class Heart(object):
     def is_alive(self):
         """Returns if the heart is beating."""
         return not (self._runner is None
-                    or not self._runner.is_alive()
-                    or self._finished.is_set())
+                    or not self._runner.is_alive())
 
     @excutils.forever_retry_uncaught_exceptions
     def _beat_forever_until_stopped(self):
         """Inner beating loop."""
-        try:
-            while not self._dead.is_set():
-                with timeutils.StopWatch() as w:
-                    wait_until_next_beat = self._driver.heartbeat()
-                ran_for = w.elapsed()
-                if ran_for > wait_until_next_beat:
-                    LOG.warning(
-                        "Heartbeating took too long to execute (it ran for"
-                        " %0.2f seconds which is %0.2f seconds longer than"
-                        " the next heartbeat idle time). This may cause"
-                        " timeouts (in locks, leadership, ...) to"
-                        " happen (which will not end well).", ran_for,
-                        ran_for - wait_until_next_beat)
-                self._beats += 1
-                # NOTE(harlowja): use the event object for waiting and
-                # not a sleep function since doing that will allow this code
-                # to terminate early if stopped via the stop() method vs
-                # having to wait until the sleep function returns.
-                self._dead.wait(wait_until_next_beat)
-        finally:
-            self._finished.set()
+        while not self._dead.is_set():
+            with timeutils.StopWatch() as w:
+                wait_until_next_beat = self._driver.heartbeat()
+            ran_for = w.elapsed()
+            has_to_sleep_for = wait_until_next_beat - ran_for
+            if has_to_sleep_for < 0:
+                LOG.warning(
+                    "Heartbeating took too long to execute (it ran for"
+                    " %0.2f seconds which is %0.2f seconds longer than"
+                    " the next heartbeat idle time). This may cause"
+                    " timeouts (in locks, leadership, ...) to"
+                    " happen (which will not end well).", ran_for,
+                    ran_for - wait_until_next_beat)
+            self._beats += 1
+            # NOTE(harlowja): use the event object for waiting and
+            # not a sleep function since doing that will allow this code
+            # to terminate early if stopped via the stop() method vs
+            # having to wait until the sleep function returns.
+            # NOTE(jd): Wait for only the half time of what we should.
+            # This is a measure of safety, better be too soon than too late.
+            self._dead.wait(has_to_sleep_for / 2.0)
 
     def start(self, thread_cls=None):
         """Starts the heart beating thread (noop if already started)."""
         if not self.is_alive():
-            self._finished.clear()
             self._dead.clear()
             self._beats = 0
             if thread_cls is None:
@@ -210,11 +221,10 @@ class Heart(object):
 
     def wait(self, timeout=None):
         """Wait up to given timeout for the heart beating thread to stop."""
-        self._finished.wait(timeout)
-        return self._finished.is_set()
+        self._runner.join(timeout)
+        return self._runner.is_alive()
 
 
-@six.add_metaclass(abc.ABCMeta)
 class CoordinationDriver(object):
 
     requires_beating = False
@@ -231,21 +241,54 @@ class CoordinationDriver(object):
     enum member(s) that can be used to interogate how this driver works.
     """
 
-    def __init__(self):
+    def __init__(self, member_id, parsed_url, options):
+        super(CoordinationDriver, self).__init__()
+        self._member_id = member_id
         self._started = False
         self._hooks_join_group = collections.defaultdict(Hooks)
         self._hooks_leave_group = collections.defaultdict(Hooks)
         self._hooks_elected_leader = collections.defaultdict(Hooks)
-        # A cache for group members
-        self._group_members = collections.defaultdict(set)
         self.requires_beating = (
             CoordinationDriver.heartbeat != self.__class__.heartbeat
         )
         self.heart = Heart(self)
 
     def _has_hooks_for_group(self, group_id):
-        return (len(self._hooks_join_group[group_id]) +
-                len(self._hooks_leave_group[group_id]))
+        return (group_id in self._hooks_join_group or
+                group_id in self._hooks_leave_group)
+
+    def join_partitioned_group(
+            self, group_id,
+            weight=1,
+            partitions=partitioner.Partitioner.DEFAULT_PARTITION_NUMBER):
+        """Join a group and get a partitioner.
+
+        A partitioner allows to distribute a bunch of objects across several
+        members using a consistent hash ring. Each object gets assigned (at
+        least) one member responsible for it. It's then possible to check which
+        object is owned by any member of the group.
+
+        This method also creates if necessary, and joins the group with the
+        selected weight.
+
+        :param group_id: The group to create a partitioner for.
+        :param weight: The weight to use in the hashring for this node.
+        :param partitions: The number of partitions to create.
+        :return: A :py:class:`~tooz.partitioner.Partitioner` object.
+
+        """
+        self.join_group_create(group_id, capabilities={'weight': weight})
+        return partitioner.Partitioner(self, group_id, partitions=partitions)
+
+    def leave_partitioned_group(self, partitioner):
+        """Leave a partitioned group.
+
+        This leaves the partitioned group and stop the partitioner.
+        :param group_id: The group to create a partitioner for.
+        """
+        leave = self.leave_group(partitioner.group_id)
+        partitioner.stop()
+        return leave.get()
 
     @staticmethod
     def run_watchers(timeout=None):
@@ -261,7 +304,6 @@ class CoordinationDriver(object):
         """Try to leader elect this coordinator & activate hooks on success."""
         raise tooz.NotImplemented
 
-    @abc.abstractmethod
     def watch_join_group(self, group_id, callback):
         """Call a function when group_id sees a new member joined.
 
@@ -274,7 +316,6 @@ class CoordinationDriver(object):
         """
         self._hooks_join_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_join_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member joined.
 
@@ -283,15 +324,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_join_group:
+                raise ValueError
             self._hooks_join_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
+        if not self._hooks_join_group[group_id]:
+            del self._hooks_join_group[group_id]
 
-    @abc.abstractmethod
     def watch_leave_group(self, group_id, callback):
         """Call a function when group_id sees a new member leaving.
 
@@ -305,7 +348,6 @@ class CoordinationDriver(object):
         """
         self._hooks_leave_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_leave_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member leaving.
 
@@ -314,15 +356,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_leave_group:
+                raise ValueError
             self._hooks_leave_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
+        if not self._hooks_leave_group[group_id]:
+            del self._hooks_leave_group[group_id]
 
-    @abc.abstractmethod
     def watch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -336,7 +380,6 @@ class CoordinationDriver(object):
         """
         self._hooks_elected_leader[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -375,12 +418,14 @@ class CoordinationDriver(object):
         is initiated.
         """
         if self._started:
-            raise ToozError(
+            raise tooz.ToozError(
                 "Can not start a driver which has not been stopped")
         self._start()
         if self.requires_beating and start_heart:
             self.heart.start()
         self._started = True
+        # Tracks which group are joined
+        self._joined_groups = set()
 
     def _start(self):
         pass
@@ -392,10 +437,24 @@ class CoordinationDriver(object):
         disappear from all joined groups.
         """
         if not self._started:
-            raise ToozError("Can not stop a driver which has not been started")
+            raise tooz.ToozError(
+                "Can not stop a driver which has not been started")
         if self.heart.is_alive():
             self.heart.stop()
             self.heart.wait()
+        # Some of the drivers modify joined_groups when being called to leave
+        # so clone it so that we aren't modifying something while iterating.
+        joined_groups = self._joined_groups.copy()
+        leaving = [self.leave_group(group) for group in joined_groups]
+        for fut in leaving:
+            try:
+                fut.get()
+            except tooz.ToozError:
+                # Whatever happens, ignore. Maybe we got booted out/never
+                # existed in the first place, or something is down, but we just
+                # want to call _stop after whatever happens to not leak any
+                # connection.
+                pass
         self._stop()
         self._started = False
 
@@ -407,7 +466,7 @@ class CoordinationDriver(object):
         """Request the creation of a group asynchronously.
 
         :param group_id: the id of the group to create
-        :type group_id: str
+        :type group_id: ascii bytes
         :returns: None
         :rtype: CoordAsyncResult
         """
@@ -427,20 +486,47 @@ class CoordinationDriver(object):
         """Join a group and establish group membership asynchronously.
 
         :param group_id: the id of the group to join
-        :type group_id: str
+        :type group_id: ascii bytes
         :param capabilities: the capabilities of the joined member
-        :type capabilities: object (typically str)
+        :type capabilities: object
         :returns: None
         :rtype: CoordAsyncResult
         """
         raise tooz.NotImplemented
 
+    @_retry.retry()
+    def join_group_create(self, group_id, capabilities=b""):
+        """Join a group and create it if necessary.
+
+        If the group cannot be joined because it does not exist, it is created
+        before being joined.
+
+        This function will keep retrying until it can create the group and join
+        it. Since nothing is transactional, it may have to retry several times
+        if another member is creating/deleting the group at the same time.
+
+        :param group_id: Identifier of the group to join and create
+        :param capabilities: the capabilities of the joined member
+        """
+        req = self.join_group(group_id, capabilities)
+        try:
+            req.get()
+        except GroupNotCreated:
+            req = self.create_group(group_id)
+            try:
+                req.get()
+            except GroupAlreadyExist:
+                # The group might have been created in the meantime, ignore
+                pass
+            # Now retry to join the group
+            raise _retry.TryAgain
+
     @staticmethod
     def leave_group(group_id):
         """Leave a group asynchronously.
 
         :param group_id: the id of the group to leave
-        :type group_id: str
+        :type group_id: ascii bytes
         :returns: None
         :rtype: CoordAsyncResult
         """
@@ -451,7 +537,7 @@ class CoordinationDriver(object):
         """Delete a group asynchronously.
 
         :param group_id: the id of the group to leave
-        :type group_id: str
+        :type group_id: ascii bytes
         :returns: Result
         :rtype: CoordAsyncResult
         """
@@ -459,9 +545,9 @@ class CoordinationDriver(object):
 
     @staticmethod
     def get_members(group_id):
-        """Return the list of all members ids of the specified group.
+        """Return the set of all members ids of the specified group.
 
-        :returns: list of all created group ids
+        :returns: set of all created group ids
         :rtype: CoordAsyncResult
         """
         raise tooz.NotImplemented
@@ -471,9 +557,9 @@ class CoordinationDriver(object):
         """Return the capabilities of a member asynchronously.
 
         :param group_id: the id of the group of the member
-        :type group_id: str
+        :type group_id: ascii bytes
         :param member_id: the id of the member
-        :type member_id: str
+        :type member_id: ascii bytes
         :returns: capabilities of a member
         :rtype: CoordAsyncResult
         """
@@ -484,9 +570,9 @@ class CoordinationDriver(object):
         """Return the statistics and capabilities of a member asynchronously.
 
         :param group_id: the id of the group of the member
-        :type group_id: str
+        :type group_id: ascii bytes
         :param member_id: the id of the member
-        :type member_id: str
+        :type member_id: ascii bytes
         :returns: capabilities and statistics of a member
         :rtype: CoordAsyncResult
         """
@@ -497,9 +583,9 @@ class CoordinationDriver(object):
         """Update member capabilities in the specified group.
 
         :param group_id: the id of the group of the current member
-        :type group_id: str
+        :type group_id: ascii bytes
         :param capabilities: the capabilities of the updated member
-        :type capabilities: object (typically str)
+        :type capabilities: object
         :returns: None
         :rtype: CoordAsyncResult
         """
@@ -550,7 +636,7 @@ class CoordAsyncResult(object):
     """
 
     @abc.abstractmethod
-    def get(self, timeout=10):
+    def get(self, timeout=None):
         """Retrieve the result of the corresponding asynchronous call.
 
         :param timeout: block until the timeout expire.
@@ -562,23 +648,107 @@ class CoordAsyncResult(object):
         """Returns True if the task is done, False otherwise."""
 
 
-class _RunWatchersMixin(object):
-    """Mixin to share the *mostly* common ``run_watchers`` implementation."""
+class CoordinatorResult(CoordAsyncResult):
+    """Asynchronous result that references a future."""
+
+    def __init__(self, fut, failure_translator=None):
+        self._fut = fut
+        self._failure_translator = failure_translator
+
+    def get(self, timeout=None):
+        try:
+            if self._failure_translator:
+                with self._failure_translator():
+                    return self._fut.result(timeout=timeout)
+            else:
+                return self._fut.result(timeout=timeout)
+        except futures.TimeoutError as e:
+            utils.raise_with_cause(OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
+
+    def done(self):
+        return self._fut.done()
+
+
+class CoordinationDriverWithExecutor(CoordinationDriver):
+
+    EXCLUDE_OPTIONS = None
+
+    def __init__(self, member_id, parsed_url, options):
+        self._options = utils.collapse(options, exclude=self.EXCLUDE_OPTIONS)
+        self._executor = utils.ProxyExecutor.build(
+            self.__class__.__name__, self._options)
+        super(CoordinationDriverWithExecutor, self).__init__(
+            member_id, parsed_url, options)
+
+    def start(self, start_heart=False):
+        self._executor.start()
+        super(CoordinationDriverWithExecutor, self).start(start_heart)
+
+    def stop(self):
+        super(CoordinationDriverWithExecutor, self).stop()
+        self._executor.stop()
+
+
+class CoordinationDriverCachedRunWatchers(CoordinationDriver):
+    """Coordination driver with a `run_watchers` implementation.
+
+    This implementation of `run_watchers` is based on a cache of the group
+    members between each run of `run_watchers` that is being updated between
+    each run.
+
+    """
+
+    def __init__(self, member_id, parsed_url, options):
+        super(CoordinationDriverCachedRunWatchers, self).__init__(
+            member_id, parsed_url, options)
+        # A cache for group members
+        self._group_members = collections.defaultdict(set)
+        self._joined_groups = set()
+
+    def _init_watch_group(self, group_id):
+        if group_id not in self._group_members:
+            members = self.get_members(group_id)
+            self._group_members[group_id] = members.get()
+
+    def watch_join_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_join_group(
+            group_id, callback)
+
+    def unwatch_join_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_join_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
+
+    def watch_leave_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_leave_group(
+            group_id, callback)
+
+    def unwatch_leave_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_leave_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
 
     def run_watchers(self, timeout=None):
         with timeutils.StopWatch(duration=timeout) as w:
-            known_groups = self.get_groups().get(
-                timeout=w.leftover(return_none=True))
             result = []
-            for group_id in known_groups:
+            group_with_hooks = set(self._hooks_join_group.keys()).union(
+                set(self._hooks_leave_group.keys()))
+            for group_id in group_with_hooks:
                 try:
-                    group_members_fut = self.get_members(group_id)
-                    group_members = group_members_fut.get(
+                    group_members = self.get_members(group_id).get(
                         timeout=w.leftover(return_none=True))
                 except GroupNotCreated:
                     group_members = set()
-                else:
-                    group_members = set(group_members)
                 if (group_id in self._joined_groups and
                         self._member_id not in group_members):
                     self._joined_groups.discard(group_id)
@@ -602,7 +772,7 @@ def get_coordinator(backend_url, member_id,
     :param backend_url: the backend URL to use
     :type backend: str
     :param member_id: the id of the member
-    :type member_id: str
+    :type member_id: ascii bytes
     :param characteristics: set
     :type characteristics: set of :py:class:`.Characteristics` that will
                            be matched to the requested driver (this **will**
@@ -642,40 +812,29 @@ def get_coordinator(backend_url, member_id,
     return d
 
 
-class ToozError(Exception):
-    """Exception raised when an internal error occurs.
-
-    Raised for instance in case of server internal error.
-
-    :ivar cause: the cause of the exception being raised, when not none this
-                 will itself be an exception instance, this is useful for
-                 creating a chain of exceptions for versions of python where
-                 this is not yet implemented/supported natively.
-
-    """
-
-    def __init__(self, message, cause=None):
-        super(ToozError, self).__init__(message)
-        self.cause = cause
+# TODO(harlowja): We'll have to figure out a way to remove this 'alias' at
+# some point in the future (when we have a better way to tell people it has
+# moved without messing up their exception catching hierarchy).
+ToozError = tooz.ToozError
 
 
-class ToozDriverChosenPoorly(ToozError):
+class ToozDriverChosenPoorly(tooz.ToozError):
     """Raised when a driver does not match desired characteristics."""
 
 
-class ToozConnectionError(ToozError):
+class ToozConnectionError(tooz.ToozError):
     """Exception raised when the client cannot connect to the server."""
 
 
-class OperationTimedOut(ToozError):
+class OperationTimedOut(tooz.ToozError):
     """Exception raised when an operation times out."""
 
 
-class LockAcquireFailed(ToozError):
+class LockAcquireFailed(tooz.ToozError):
     """Exception raised when a lock acquire fails in a context manager."""
 
 
-class GroupNotCreated(ToozError):
+class GroupNotCreated(tooz.ToozError):
     """Exception raised when the caller request an nonexistent group."""
     def __init__(self, group_id):
         self.group_id = group_id
@@ -683,7 +842,7 @@ class GroupNotCreated(ToozError):
             "Group %s does not exist" % group_id)
 
 
-class GroupAlreadyExist(ToozError):
+class GroupAlreadyExist(tooz.ToozError):
     """Exception raised trying to create an already existing group."""
     def __init__(self, group_id):
         self.group_id = group_id
@@ -691,7 +850,7 @@ class GroupAlreadyExist(ToozError):
             "Group %s already exists" % group_id)
 
 
-class MemberAlreadyExist(ToozError):
+class MemberAlreadyExist(tooz.ToozError):
     """Exception raised trying to join a group already joined."""
     def __init__(self, group_id, member_id):
         self.group_id = group_id
@@ -701,7 +860,7 @@ class MemberAlreadyExist(ToozError):
             (member_id, group_id))
 
 
-class MemberNotJoined(ToozError):
+class MemberNotJoined(tooz.ToozError):
     """Exception raised trying to access a member not in a group."""
     def __init__(self, group_id, member_id):
         self.group_id = group_id
@@ -710,14 +869,14 @@ class MemberNotJoined(ToozError):
                                               (member_id, group_id))
 
 
-class GroupNotEmpty(ToozError):
+class GroupNotEmpty(tooz.ToozError):
     "Exception raised when the caller try to delete a group with members."
     def __init__(self, group_id):
         self.group_id = group_id
         super(GroupNotEmpty, self).__init__("Group %s is not empty" % group_id)
 
 
-class WatchCallbackNotFound(ToozError):
+class WatchCallbackNotFound(tooz.ToozError):
     """Exception raised when unwatching a group.
 
     Raised when the caller tries to unwatch a group with a callback that
@@ -732,29 +891,7 @@ class WatchCallbackNotFound(ToozError):
             (callback.__name__, group_id))
 
 
-class SerializationError(ToozError):
-    "Exception raised when serialization or deserialization breaks."
-
-
-def raise_with_cause(exc_cls, message, *args, **kwargs):
-    """Helper to raise + chain exceptions (when able) and associate a *cause*.
-
-    **For internal usage only.**
-
-    NOTE(harlowja): Since in py3.x exceptions can be chained (due to
-    :pep:`3134`) we should try to raise the desired exception with the given
-    *cause*.
-
-    :param exc_cls: the :py:class:`~tooz.coordination.ToozError` class
-                    to raise.
-    :param message: the text/str message that will be passed to
-                    the exceptions constructor as its first positional
-                    argument.
-    :param args: any additional positional arguments to pass to the
-                 exceptions constructor.
-    :param kwargs: any additional keyword arguments to pass to the
-                   exceptions constructor.
-    """
-    if not issubclass(exc_cls, ToozError):
-        raise ValueError("Subclass of tooz error is required")
-    excutils.raise_with_cause(exc_cls, message, *args, **kwargs)
+# TODO(harlowja,jd): We'll have to figure out a way to remove this 'alias' at
+# some point in the future (when we have a better way to tell people it has
+# moved without messing up their exception catching hierarchy).
+SerializationError = utils.SerializationError
