@@ -19,8 +19,8 @@ from oslo_utils import encodeutils
 import pymysql
 
 import tooz
+from tooz import _retry
 from tooz import coordination
-from tooz.drivers import _retry
 from tooz import locking
 from tooz import utils
 
@@ -35,9 +35,12 @@ class MySQLLock(locking.Lock):
     def __init__(self, name, parsed_url, options):
         super(MySQLLock, self).__init__(name)
         self.acquired = False
-        self._conn = MySQLDriver.get_connection(parsed_url, options)
+        self._conn = MySQLDriver.get_connection(parsed_url, options, True)
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+
+        if shared:
+            raise tooz.NotImplemented
 
         @_retry.retry(stop_max_delay=blocking)
         def _lock():
@@ -52,10 +55,12 @@ class MySQLLock(locking.Lock):
             # So, we track locally the lock status with self.acquired
             if self.acquired is True:
                 if blocking:
-                    raise _retry.Retry
+                    raise _retry.TryAgain
                 return False
 
             try:
+                if not self._conn.open:
+                    self._conn.connect()
                 with self._conn as cur:
                     cur.execute("SELECT GET_LOCK(%s, 0);", self.name)
                     # Can return NULL on error
@@ -63,16 +68,23 @@ class MySQLLock(locking.Lock):
                         self.acquired = True
                         return True
             except pymysql.MySQLError as e:
-                coordination.raise_with_cause(
-                    coordination.ToozError,
+                utils.raise_with_cause(
+                    tooz.ToozError,
                     encodeutils.exception_to_unicode(e),
                     cause=e)
 
             if blocking:
-                raise _retry.Retry
+                raise _retry.TryAgain
+            self._conn.close()
             return False
 
-        return _lock()
+        try:
+            return _lock()
+        except Exception:
+            # Close the connection if we tried too much and finally failed, or
+            # anything else bad happened.
+            self._conn.close()
+            raise
 
     def release(self):
         if not self.acquired:
@@ -81,12 +93,13 @@ class MySQLLock(locking.Lock):
             with self._conn as cur:
                 cur.execute("SELECT RELEASE_LOCK(%s);", self.name)
                 cur.fetchone()
-                self.acquired = False
-                return True
+            self.acquired = False
+            self._conn.close()
+            return True
         except pymysql.MySQLError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def __del__(self):
         if self.acquired:
@@ -117,7 +130,7 @@ class MySQLDriver(coordination.CoordinationDriver):
 
     def __init__(self, member_id, parsed_url, options):
         """Initialize the MySQL driver."""
-        super(MySQLDriver, self).__init__()
+        super(MySQLDriver, self).__init__(member_id, parsed_url, options)
         self._parsed_url = parsed_url
         self._options = utils.collapse(options)
 
@@ -156,7 +169,7 @@ class MySQLDriver(coordination.CoordinationDriver):
         raise tooz.NotImplemented
 
     @staticmethod
-    def get_connection(parsed_url, options):
+    def get_connection(parsed_url, options, defer_connect=False):
         host = parsed_url.hostname
         port = parsed_url.port or MySQLLock.MYSQL_DEFAULT_PORT
         dbname = parsed_url.path[1:]
@@ -170,14 +183,16 @@ class MySQLDriver(coordination.CoordinationDriver):
                                        port=port,
                                        user=username,
                                        passwd=password,
-                                       database=dbname)
+                                       database=dbname,
+                                       defer_connect=defer_connect)
             else:
                 return pymysql.Connect(host=host,
                                        port=port,
                                        user=username,
                                        passwd=password,
-                                       database=dbname)
+                                       database=dbname,
+                                       defer_connect=defer_connect)
         except (pymysql.err.OperationalError, pymysql.err.InternalError) as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
