@@ -23,8 +23,8 @@ import psycopg2
 import six
 
 import tooz
+from tooz import _retry
 from tooz import coordination
-from tooz.drivers import _retry
 from tooz import locking
 from tooz import utils
 
@@ -85,9 +85,9 @@ def _translating_cursor(conn):
         with conn.cursor() as cur:
             yield cur
     except psycopg2.Error as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      _format_exception(e),
-                                      cause=e)
+        utils.raise_with_cause(tooz.ToozError,
+                               _format_exception(e),
+                               cause=e)
 
 
 class PostgresLock(locking.Lock):
@@ -96,7 +96,9 @@ class PostgresLock(locking.Lock):
     def __init__(self, name, parsed_url, options):
         super(PostgresLock, self).__init__(name)
         self.acquired = False
-        self._conn = PostgresDriver.get_connection(parsed_url, options)
+        self._conn = None
+        self._parsed_url = parsed_url
+        self._options = options
         h = hashlib.md5()
         h.update(name)
         if six.PY2:
@@ -104,7 +106,10 @@ class PostgresLock(locking.Lock):
         else:
             self.key = h.digest()[0:2]
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+
+        if shared:
+            raise tooz.NotImplemented
 
         @_retry.retry(stop_max_delay=blocking)
         def _lock():
@@ -112,12 +117,17 @@ class PostgresLock(locking.Lock):
             # so we track it internally if the process already has the lock.
             if self.acquired is True:
                 if blocking:
-                    raise _retry.Retry
+                    raise _retry.TryAgain
                 return False
 
+            if not self._conn or self._conn.closed:
+                self._conn = PostgresDriver.get_connection(self._parsed_url,
+                                                           self._options)
+
             with _translating_cursor(self._conn) as cur:
                 if blocking is True:
-                    cur.execute("SELECT pg_advisory_lock(%s, %s);", self.key)
+                    cur.execute("SELECT pg_advisory_lock(%s, %s);",
+                                self.key)
                     cur.fetchone()
                     self.acquired = True
                     return True
@@ -128,25 +138,32 @@ class PostgresLock(locking.Lock):
                         self.acquired = True
                         return True
                     elif blocking is False:
+                        self._conn.close()
                         return False
                     else:
-                        raise _retry.Retry
+                        raise _retry.TryAgain
 
-        return _lock()
+        try:
+            return _lock()
+        except Exception:
+            if self._conn:
+                self._conn.close()
+            raise
 
     def release(self):
         if not self.acquired:
             return False
+
         with _translating_cursor(self._conn) as cur:
-            cur.execute("SELECT pg_advisory_unlock(%s, %s);",
-                        self.key)
+            cur.execute("SELECT pg_advisory_unlock(%s, %s);", self.key)
             cur.fetchone()
-            self.acquired = False
-            return True
+        self.acquired = False
+        self._conn.close()
+        return True
 
     def __del__(self):
         if self.acquired:
-            LOG.warn("unreleased lock %s garbage collected" % self.name)
+            LOG.warning("unreleased lock %s garbage collected", self.name)
 
 
 class PostgresDriver(coordination.CoordinationDriver):
@@ -161,6 +178,7 @@ class PostgresDriver(coordination.CoordinationDriver):
     """
 
     CHARACTERISTICS = (
+        coordination.Characteristics.NON_TIMEOUT_BASED,
         coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
         coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
         coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
@@ -172,13 +190,12 @@ class PostgresDriver(coordination.CoordinationDriver):
 
     def __init__(self, member_id, parsed_url, options):
         """Initialize the PostgreSQL driver."""
-        super(PostgresDriver, self).__init__()
+        super(PostgresDriver, self).__init__(member_id, parsed_url, options)
         self._parsed_url = parsed_url
         self._options = utils.collapse(options)
 
     def _start(self):
-        self._conn = PostgresDriver.get_connection(self._parsed_url,
-                                                   self._options)
+        self._conn = self.get_connection(self._parsed_url, self._options)
 
     def _stop(self):
         self._conn.close()
@@ -215,16 +232,18 @@ class PostgresDriver(coordination.CoordinationDriver):
         host = options.get("host") or parsed_url.hostname
         port = options.get("port") or parsed_url.port
         dbname = options.get("dbname") or parsed_url.path[1:]
-        username = parsed_url.username
-        password = parsed_url.password
+        kwargs = {}
+        if parsed_url.username is not None:
+            kwargs["user"] = parsed_url.username
+        if parsed_url.password is not None:
+            kwargs["password"] = parsed_url.password
 
         try:
             return psycopg2.connect(host=host,
                                     port=port,
-                                    user=username,
-                                    password=password,
-                                    database=dbname)
+                                    database=dbname,
+                                    **kwargs)
         except psycopg2.Error as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          _format_exception(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   _format_exception(e),
+                                   cause=e)
