@@ -18,10 +18,11 @@ from __future__ import absolute_import
 
 import contextlib
 from distutils import version
+import functools
 import logging
 import string
+import threading
 
-from concurrent import futures
 from oslo_utils import encodeutils
 from oslo_utils import strutils
 import redis
@@ -45,19 +46,22 @@ def _translate_failures():
     try:
         yield
     except (exceptions.ConnectionError, exceptions.TimeoutError) as e:
-        coordination.raise_with_cause(coordination.ToozConnectionError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
     except exceptions.RedisError as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
 
 class RedisLock(locking.Lock):
     def __init__(self, coord, client, name, timeout):
         name = "%s_%s_lock" % (coord.namespace, six.text_type(name))
         super(RedisLock, self).__init__(name)
+        # NOTE(jd) Make sure we don't release and heartbeat at the same time by
+        # using a exclusive access lock (LP#1557593)
+        self._exclusive_access = threading.Lock()
         self._lock = client.lock(name,
                                  timeout=timeout,
                                  thread_local=False)
@@ -76,38 +80,47 @@ class RedisLock(locking.Lock):
         with _translate_failures():
             return bool(self._client.delete(self.name))
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
         blocking, timeout = utils.convert_blocking(blocking)
         with _translate_failures():
             acquired = self._lock.acquire(
                 blocking=blocking, blocking_timeout=timeout)
             if acquired:
-                self._coord._acquired_locks.add(self)
+                with self._exclusive_access:
+                    self._coord._acquired_locks.add(self)
             return acquired
 
     def release(self):
-        if not self.acquired:
-            return False
-        with _translate_failures():
-            try:
-                self._lock.release()
-            except exceptions.LockError:
+        with self._exclusive_access:
+            if not self.acquired:
                 return False
-            self._coord._acquired_locks.discard(self)
-            return True
+            with _translate_failures():
+                try:
+                    self._lock.release()
+                except exceptions.LockError as e:
+                    LOG.error("Unable to release lock '%r': %s", self, e)
+                    return False
+                finally:
+                    self._coord._acquired_locks.discard(self)
+                return True
 
     def heartbeat(self):
-        if self.acquired:
-            with _translate_failures():
-                self._lock.extend(self._lock.timeout)
+        with self._exclusive_access:
+            if self.acquired:
+                with _translate_failures():
+                    self._lock.extend(self._lock.timeout)
+                    return True
+        return False
 
     @property
     def acquired(self):
         return self in self._coord._acquired_locks
 
 
-class RedisDriver(coordination._RunWatchersMixin,
-                  coordination.CoordinationDriver):
+class RedisDriver(coordination.CoordinationDriverCachedRunWatchers,
+                  coordination.CoordinationDriverWithExecutor):
     """Redis provides a few nice benefits that act as a poormans zookeeper.
 
     It **is** fully functional and implements all of the coordination
@@ -315,28 +328,25 @@ return 1
     .. _Lua: http://www.lua.org
     """
 
+    EXCLUDE_OPTIONS = CLIENT_LIST_ARGS
+
     def __init__(self, member_id, parsed_url, options):
-        super(RedisDriver, self).__init__()
-        options = utils.collapse(options, exclude=self.CLIENT_LIST_ARGS)
+        super(RedisDriver, self).__init__(member_id, parsed_url, options)
         self._parsed_url = parsed_url
-        self._options = options
-        self._encoding = options.get('encoding', self.DEFAULT_ENCODING)
-        timeout = options.get('timeout', self.CLIENT_DEFAULT_SOCKET_TO)
+        self._encoding = self._options.get('encoding', self.DEFAULT_ENCODING)
+        timeout = self._options.get('timeout', self.CLIENT_DEFAULT_SOCKET_TO)
         self.timeout = int(timeout)
-        self.membership_timeout = float(options.get(
+        self.membership_timeout = float(self._options.get(
             'membership_timeout', timeout))
-        lock_timeout = options.get('lock_timeout', self.timeout)
+        lock_timeout = self._options.get('lock_timeout', self.timeout)
         self.lock_timeout = int(lock_timeout)
-        namespace = options.get('namespace', self.DEFAULT_NAMESPACE)
+        namespace = self._options.get('namespace', self.DEFAULT_NAMESPACE)
         self._namespace = utils.to_binary(namespace, encoding=self._encoding)
         self._group_prefix = self._namespace + b"_group"
         self._beat_prefix = self._namespace + b"_beats"
         self._groups = self._namespace + b"_groups"
         self._client = None
-        self._member_id = utils.to_binary(member_id, encoding=self._encoding)
         self._acquired_locks = set()
-        self._joined_groups = set()
-        self._executor = utils.ProxyExecutor.build("Redis", options)
         self._started = False
         self._server_info = {}
         self._scripts = {}
@@ -423,14 +433,14 @@ return 1
         return redis.StrictRedis(**kwargs)
 
     def _start(self):
-        self._executor.start()
+        super(RedisDriver, self)._start()
         try:
             self._client = self._make_client(self._parsed_url, self._options,
                                              self.timeout)
         except exceptions.RedisError as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             # Ensure that the server is alive and not dead, this does not
             # ensure the server will always be alive, but does insure that it
@@ -504,7 +514,7 @@ return 1
         for lock in self._acquired_locks.copy():
             try:
                 lock.heartbeat()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to heartbeat lock '%s'", lock,
                             exc_info=True)
         return min(self.lock_timeout, self.membership_timeout)
@@ -514,19 +524,9 @@ return 1
             lock = self._acquired_locks.pop()
             try:
                 lock.release()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to release lock '%s'", lock, exc_info=True)
-        while self._joined_groups:
-            group_id = self._joined_groups.pop()
-            try:
-                self.leave_group(group_id).get()
-            except (coordination.MemberNotJoined,
-                    coordination.GroupNotCreated):
-                pass
-            except coordination.ToozError:
-                LOG.warning("Unable to leave group '%s'", group_id,
-                            exc_info=True)
-        self._executor.stop()
+        super(RedisDriver, self)._stop()
         if self._client is not None:
             # Make sure we no longer exist...
             beat_id = self._encode_beat_id(self._member_id)
@@ -535,7 +535,7 @@ return 1
                 # exist in the first place, which is fine/expected/desired...
                 with _translate_failures():
                     self._client.delete(beat_id)
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to delete heartbeat key '%s'", beat_id,
                             exc_info=True)
             self._client = None
@@ -545,14 +545,14 @@ return 1
 
     def _submit(self, cb, *args, **kwargs):
         if not self._started:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
         return self._executor.submit(cb, *args, **kwargs)
 
     def _get_script(self, script_key):
         try:
             return self._scripts[script_key]
         except KeyError:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
 
     def create_group(self, group_id):
         script = self._get_script('create_group')
@@ -617,13 +617,13 @@ return 1
         def _get_members(p):
             if not p.exists(encoded_group):
                 raise coordination.GroupNotCreated(group_id)
-            potential_members = []
+            potential_members = set()
             for m in p.hkeys(encoded_group):
                 m = self._decode_member_id(m)
                 if m != self.GROUP_EXISTS:
-                    potential_members.append(m)
+                    potential_members.add(m)
             if not potential_members:
-                return []
+                return set()
             # Ok now we need to see which members have passed away...
             gone_members = set()
             member_values = p.mget(compat_map(self._encode_beat_id,
@@ -644,10 +644,9 @@ return 1
                                             for m in gone_members)
                 p.hdel(encoded_group, *encoded_gone_members)
                 p.execute()
-                return list(m for m in potential_members
-                            if m not in gone_members)
-            else:
-                return potential_members
+                return set(m for m in potential_members
+                           if m not in gone_members)
+            return potential_members
 
         return RedisFutureResult(self._submit(self._client.transaction,
                                               _get_members, encoded_group,
@@ -710,13 +709,13 @@ return 1
             if result == -3:
                 raise coordination.GroupNotEmpty(group_id)
             if result == -4:
-                raise coordination.ToozError("Unable to remove '%s' key"
-                                             " from set located at '%s'"
-                                             % (args[0], keys[-1]))
+                raise tooz.ToozError("Unable to remove '%s' key"
+                                     " from set located at '%s'"
+                                     % (args[0], keys[-1]))
             if result != 1:
-                raise coordination.ToozError("Internal error, unable"
-                                             " to complete group '%s' removal"
-                                             % (group_id))
+                raise tooz.ToozError("Internal error, unable"
+                                     " to complete group '%s' removal"
+                                     % (group_id))
 
         return RedisFutureResult(self._submit(_delete_group, script))
 
@@ -734,32 +733,6 @@ return 1
 
         return RedisFutureResult(self._submit(_get_groups))
 
-    def _init_watch_group(self, group_id):
-        members = self.get_members(group_id)
-        self._group_members[group_id].update(members.get(timeout=None))
-
-    def watch_join_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(RedisDriver, self).watch_join_group(group_id, callback)
-
-    def unwatch_join_group(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_join_group(group_id, callback)
-
-    def watch_leave_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(RedisDriver, self).watch_leave_group(group_id, callback)
-
-    def unwatch_leave_group(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_leave_group(group_id, callback)
-
-    def watch_elected_as_leader(self, group_id, callback):
-        return super(RedisDriver, self).watch_elected_as_leader(
-            group_id, callback)
-
-    def unwatch_elected_as_leader(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_elected_as_leader(
-            group_id, callback)
-
     def _get_leader_lock(self, group_id):
         name = self._encode_group_leader(group_id)
         return self.get_lock(name)
@@ -778,24 +751,5 @@ return 1
         return result
 
 
-class RedisFutureResult(coordination.CoordAsyncResult):
-    """Redis asynchronous result that references a future."""
-
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=10):
-        try:
-            # Late translate the common failures since the redis client
-            # may throw things that we can not catch in the callbacks where
-            # it is used (especially one that uses the transaction
-            # method).
-            with _translate_failures():
-                return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
-
-    def done(self):
-        return self._fut.done()
+RedisFutureResult = functools.partial(coordination.CoordinatorResult,
+                                      failure_translator=_translate_failures)
